"""
Remote clipboard client and server package.

The package exposes CLI entrypoints:
- remoclip: client CLI
- remoclip_server: server CLI
"""

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("remoclip")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "0.0.0"

__all__ = ["__version__"]



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Literal, Mapping

import yaml

DEFAULT_CONFIG_PATH = Path("~/.remoclip.yaml").expanduser()

SECURITY_TOKEN_HEADER = "X-RemoClip-Token"


ClipboardBackendName = Literal["system", "private"]


DEFAULT_CONFIG: dict[str, Any] = {
    "security_token": None,
    "server": {
        "host": "127.0.0.1",
        "port": 35612,
        "db": "~/.remoclip.sqlite",
        "clipboard_backend": "system",
        "allow_deletions": False,
    },
    "client": {
        "url": "http://127.0.0.1:35612",
        "socket": None,
    },
}


@dataclass(frozen=True)
class ServerConfig:
    host: str
    port: int
    db: Path
    clipboard_backend: ClipboardBackendName = "system"
    allow_deletions: bool = False

    @property
    def db_path(self) -> Path:
        return self.db.expanduser()


@dataclass(frozen=True)
class ClientConfig:
    url: str
    socket: Path | None = None

    @property
    def socket_path(self) -> Path | None:
        if self.socket is None:
            return None
        return self.socket.expanduser()


@dataclass(frozen=True)
class RemoClipConfig:
    security_token: str | None
    server: ServerConfig
    client: ClientConfig


def load_config(path: str | None = None) -> RemoClipConfig:
    """Load configuration from YAML file, filling in defaults."""
    config_path = Path(path).expanduser() if path else DEFAULT_CONFIG_PATH
    data = _merge(DEFAULT_CONFIG, _load_yaml(config_path))

    server_config = data["server"]
    client_config = data["client"]

    server = ServerConfig(
        host=str(server_config["host"]),
        port=int(server_config["port"]),
        db=Path(str(server_config["db"])),
        clipboard_backend=_normalize_clipboard_backend(
            server_config.get("clipboard_backend")
        ),
        allow_deletions=_normalize_allow_deletions(server_config.get("allow_deletions")),
    )

    socket_value = client_config.get("socket")
    if socket_value in (None, ""):
        socket_path = None
    else:
        socket_path = Path(str(socket_value))

    client = ClientConfig(
        url=str(client_config["url"]),
        socket=socket_path,
    )

    security_token = data.get("security_token")
    if security_token is not None:
        security_token = str(security_token)

    return RemoClipConfig(
        security_token=security_token,
        server=server,
        client=client,
    )


def _normalize_clipboard_backend(value: Any) -> ClipboardBackendName:
    backend = str(value or "system").lower()
    if backend not in ("system", "private"):
        raise ValueError(
            "clipboard_backend must be either 'system' or 'private'"
        )
    return backend  # type: ignore[return-value]


def _normalize_allow_deletions(value: Any | None) -> bool:
    if value is None:
        return False
    if isinstance(value, bool):
        return value
    raise TypeError("allow_deletions must be a boolean")


def _merge(defaults: Mapping[str, Any], overrides: Mapping[str, Any] | None) -> dict[str, Any]:
    if overrides is None:
        return {key: _clone(value) for key, value in defaults.items()}

    merged: dict[str, Any] = {}
    for key, default_value in defaults.items():
        if key not in overrides or overrides[key] is None:
            merged[key] = _clone(default_value)
            continue
        override_value = overrides[key]
        if isinstance(default_value, Mapping) and isinstance(override_value, Mapping):
            merged[key] = _merge(default_value, override_value)
        else:
            merged[key] = override_value
    for key, value in overrides.items():
        if key not in merged and value is not None:
            merged[key] = value
    return merged


def _clone(value: Any) -> Any:
    if isinstance(value, Mapping):
        return {k: _clone(v) for k, v in value.items()}
    return value


def _load_yaml(path: Path) -> Mapping[str, Any] | None:
    if not path.exists():
        return None
    loaded = yaml.safe_load(path.read_text())
    if loaded is None:
        return None
    if not isinstance(loaded, Mapping):
        raise TypeError("Configuration file must contain a mapping at the top level")
    return loaded



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Protocol

try:  # pragma: no cover - import guard
    import pyperclip  # type: ignore
except ModuleNotFoundError:  # pragma: no cover - import guard
    pyperclip = None  # type: ignore[assignment]


class ClipboardBackend(Protocol):
    """Minimal interface for clipboard implementations."""

    def copy(self, text: str) -> None:
        """Persist *text* to the clipboard."""

    def paste(self) -> str:
        """Return the last persisted clipboard value."""


@dataclass
class PrivateClipboardBackend:
    """In-process clipboard implementation used for headless deployments."""

    _value: str = ""

    def copy(self, text: str) -> None:
        self._value = text

    def paste(self) -> str:
        return self._value


class SystemClipboardBackend:
    """Wrapper around :mod:`pyperclip` for system clipboard access."""

    def __init__(self) -> None:
        if pyperclip is None:
            raise RuntimeError("pyperclip is not available")

    def copy(self, text: str) -> None:
        assert pyperclip is not None  # for type checkers
        pyperclip.copy(text)

    def paste(self) -> str:
        assert pyperclip is not None  # for type checkers
        return str(pyperclip.paste())


def is_system_clipboard_available() -> bool:
    """Return ``True`` when the system clipboard backend can be constructed."""

    return pyperclip is not None


def warn_if_unavailable(logger: logging.Logger, backend: str) -> None:
    """Emit a warning that *backend* is unavailable."""

    logger.warning(
        "%s clipboard backend selected but pyperclip is not available; falling back to private backend",
        backend,
    )



from __future__ import annotations

from contextlib import contextmanager
from datetime import datetime, timezone
from pathlib import Path
from typing import Iterator

from sqlalchemy import Column, DateTime, Integer, String, Text, create_engine
from sqlalchemy.orm import declarative_base, sessionmaker, Session

Base = declarative_base()


def utc_now() -> datetime:
    return datetime.now(timezone.utc)


class ClipboardEvent(Base):
    __tablename__ = "clipboard_events"

    id = Column(Integer, primary_key=True)
    timestamp = Column(DateTime(timezone=True), default=utc_now, nullable=False)
    hostname = Column(String(255), nullable=False)
    action = Column(String(32), nullable=False)
    content = Column(Text, nullable=False)


def ensure_directory(path: Path) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)


def create_session_factory(db_path: Path):
    ensure_directory(db_path)
    engine = create_engine(
        f"sqlite:///{db_path}",
        connect_args={"check_same_thread": False},
        future=True,
    )
    Base.metadata.create_all(engine)
    return sessionmaker(bind=engine, autoflush=False, autocommit=False, future=True)


@contextmanager
def session_scope(session_factory) -> Iterator[Session]:
    session: Session | None = None
    try:
        session = session_factory()
        yield session
        session.commit()
    except Exception:
        if session is not None:
            session.rollback()
        raise
    finally:
        if session is not None:
            session.close()



from __future__ import annotations

import argparse
import json
import socket
import sys
from http.client import HTTPConnection, HTTPException
from pathlib import Path
from typing import Any

from urllib.parse import quote, urlsplit

import requests
from requests import Response, Session as RequestsSession

from .config import (
    DEFAULT_CONFIG_PATH,
    SECURITY_TOKEN_HEADER,
    RemoClipConfig,
    load_config,
)


class _UnixSocketHTTPConnection(HTTPConnection):
    def __init__(self, socket_path: str, timeout: float | None):
        super().__init__("localhost", timeout=timeout)
        self._socket_path = socket_path

    def connect(self) -> None:  # pragma: no cover - exercised indirectly by session
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        if self.timeout is not None:
            sock.settimeout(self.timeout)
        sock.connect(self._socket_path)
        self.sock = sock


class _UnixSocketResponse:
    def __init__(self, status_code: int, payload: dict[str, Any]):
        self.status_code = status_code
        self._payload = payload

    def raise_for_status(self) -> None:
        if self.status_code >= 400:
            raise requests.HTTPError(f"{self.status_code}")

    def json(self) -> dict[str, Any]:
        return self._payload


class UnixSocketSession:
    def __init__(self, socket_path: Path):
        self._socket_path = str(socket_path)

    def _request(
        self,
        method: str,
        url: str,
        *,
        json_payload: dict[str, Any] | None,
        headers: dict[str, str] | None,
        timeout: float,
    ) -> _UnixSocketResponse:
        body: bytes | None = None
        request_headers = dict(headers or {})
        if json_payload is not None:
            body = json.dumps(json_payload).encode("utf-8")
            request_headers.setdefault("Content-Type", "application/json")
        parsed = urlsplit(url)
        request_path = parsed.path or "/"
        if parsed.query:
            request_path = f"{request_path}?{parsed.query}"
        connection = _UnixSocketHTTPConnection(self._socket_path, timeout)
        try:
            connection.request(method.upper(), request_path, body=body, headers=request_headers)
            response = connection.getresponse()
            raw_data = response.read()
            status = response.status
            reason = response.reason
            headers_map = dict(response.getheaders())
        except (OSError, HTTPException) as exc:  # pragma: no cover - connection issues are exceptional
            raise requests.RequestException(str(exc)) from exc
        finally:
            connection.close()

        if status >= 400:
            http_response = Response()
            http_response.status_code = status
            http_response.reason = reason
            http_response.headers = headers_map
            http_response._content = raw_data
            raise requests.HTTPError(f"{status} {reason}", response=http_response)

        payload: dict[str, Any]
        if raw_data:
            payload = json.loads(raw_data.decode("utf-8"))
        else:
            payload = {}
        return _UnixSocketResponse(status, payload)

    def post(
        self,
        url: str,
        *,
        json: dict[str, Any],
        headers: dict[str, str] | None,
        timeout: float,
    ) -> _UnixSocketResponse:
        return self._request(
            "POST",
            url,
            json_payload=json,
            headers=headers,
            timeout=timeout,
        )

    def get(
        self,
        url: str,
        *,
        json: dict[str, Any],
        headers: dict[str, str] | None,
        timeout: float,
    ) -> _UnixSocketResponse:
        return self._request(
            "GET",
            url,
            json_payload=json,
            headers=headers,
            timeout=timeout,
        )

    def delete(
        self,
        url: str,
        *,
        json: dict[str, Any],
        headers: dict[str, str] | None,
        timeout: float,
    ) -> _UnixSocketResponse:
        return self._request(
            "DELETE",
            url,
            json_payload=json,
            headers=headers,
            timeout=timeout,
        )


class RemoClipClient:
    def __init__(self, config: RemoClipConfig):
        self.config = config
        socket_path = config.client.socket_path
        if socket_path is not None:
            encoded_path = quote(str(socket_path), safe="")
            self.base_url = f"http+unix://{encoded_path}"
            session = UnixSocketSession(socket_path)
        else:
            self.base_url = config.client.url.rstrip("/")
            session = RequestsSession()
        self._session = session
        self._headers = {}
        if config.security_token:
            self._headers[SECURITY_TOKEN_HEADER] = config.security_token

    def _payload(self, extra: dict[str, Any] | None = None) -> dict[str, Any]:
        payload = {"hostname": socket.gethostname()}
        if extra:
            payload.update(extra)
        return payload

    def copy(self, content: str, timeout: float = 5.0) -> dict[str, Any]:
        response = self._session.post(
            f"{self.base_url}/copy",
            json=self._payload({"content": content}),
            headers=self._headers,
            timeout=timeout,
        )
        response.raise_for_status()
        return response.json()

    def paste(self, event_id: int | None = None, timeout: float = 5.0) -> str:
        extra: dict[str, Any] | None = None
        if event_id is not None:
            extra = {"id": event_id}
        response = self._session.get(
            f"{self.base_url}/paste",
            json=self._payload(extra),
            headers=self._headers,
            timeout=timeout,
        )
        response.raise_for_status()
        data = response.json()
        return data.get("content", "")

    def history(
        self,
        limit: int | None = None,
        event_id: int | None = None,
        timeout: float = 5.0,
    ) -> dict[str, Any]:
        extra: dict[str, Any] = {}
        if limit is not None:
            extra["limit"] = limit
        if event_id is not None:
            extra["id"] = event_id
        response = self._session.get(
            f"{self.base_url}/history",
            json=self._payload(extra),
            headers=self._headers,
            timeout=timeout,
        )
        response.raise_for_status()
        return response.json()

    def delete_history(self, event_id: int, timeout: float = 5.0) -> dict[str, Any]:
        response = self._session.delete(
            f"{self.base_url}/history",
            json=self._payload({"id": event_id}),
            headers=self._headers,
            timeout=timeout,
        )
        response.raise_for_status()
        return response.json()


def main() -> None:
    parser = argparse.ArgumentParser(description="remoclip client CLI")
    parser.add_argument(
        "--config",
        default=str(DEFAULT_CONFIG_PATH),
        help="Path to configuration file (default: ~/.remoclip.yaml)",
    )
    parser.add_argument(
        "command",
        choices=["copy", "c", "paste", "p", "history", "h"],
        help="Action to perform on the remote clipboard",
    )
    parser.add_argument(
        "--limit",
        type=int,
        help="Limit for history entries (only for history command)",
    )
    parser.add_argument(
        "--id",
        type=int,
        help="Retrieve a specific entry by id (available for paste and history commands)",
    )
    parser.add_argument(
        "--delete",
        action="store_true",
        help="Delete a specific history entry by id (history command only)",
    )
    parser.add_argument(
        "-s",
        "--strip",
        action="store_true",
        help=(
            "Remove trailing newline characters from stdin before copying "
            "(copy command only)"
        ),
    )

    args = parser.parse_args()
    config = load_config(args.config)
    client = RemoClipClient(config)

    try:
        if args.strip and args.command not in ("copy", "c"):
            raise ValueError("--strip can only be used with the copy command")
        if args.command in ("copy", "c"):
            content = sys.stdin.read()
            if args.strip:
                content = content.rstrip("\n")
            client.copy(content)
            sys.stdout.write(content)
        elif args.command in ("paste", "p"):
            if args.id is not None and args.id <= 0:
                raise ValueError("id must be a positive integer")
            content = client.paste(event_id=args.id)
            sys.stdout.write(content)
        elif args.command in ("history", "h"):
            if args.delete:
                if args.id is None:
                    raise ValueError("id must be provided when deleting a history entry")
                if args.id <= 0:
                    raise ValueError("id must be a positive integer")
                if args.limit is not None:
                    raise ValueError("limit cannot be combined with --delete")
                result = client.delete_history(event_id=args.id)
                json.dump(result, sys.stdout, indent=2)
                sys.stdout.write("\n")
            else:
                if args.limit is not None and args.limit <= 0:
                    raise ValueError("limit must be a positive integer")
                if args.id is not None and args.id <= 0:
                    raise ValueError("id must be a positive integer")
                history = client.history(limit=args.limit, event_id=args.id)
                json.dump(history, sys.stdout, indent=2)
                sys.stdout.write("\n")
    except requests.RequestException as exc:
        sys.stderr.write(f"Request failed: {exc}\n")
        sys.exit(1)
    except ValueError as exc:
        sys.stderr.write(f"Error: {exc}\n")
        sys.exit(2)


if __name__ == "__main__":  # pragma: no cover
    main()



from __future__ import annotations

import argparse
import json
import logging
from typing import Any
from datetime import datetime, timezone
from werkzeug.serving import WSGIRequestHandler, make_server

from flask import Flask, jsonify, request

from .config import (
    DEFAULT_CONFIG_PATH,
    SECURITY_TOKEN_HEADER,
    RemoClipConfig,
    load_config,
)
from .clipboard import (
    ClipboardBackend,
    PrivateClipboardBackend,
    SystemClipboardBackend,
    is_system_clipboard_available,
    warn_if_unavailable,
)
from .db import ClipboardEvent, create_session_factory, session_scope


class LoggingWSGIRequestHandler(WSGIRequestHandler):
    """WSGI request handler that forwards access logs to :mod:`logging`."""

    level_map = {
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
    }

    def log(self, type: str, message: str, *args: Any) -> None:  # pragma: no cover - IO heavy
        logger = logging.getLogger("werkzeug.server")
        level = self.level_map.get(type, logging.INFO)
        formatted = message % args if args else message
        logger.log(
            level,
            "%s - - [%s] %s",
            self.address_string(),
            self.log_date_time_string(),
            formatted,
        )

    def log_request(
        self, code: int | str = "-", size: int | str = "-"
    ) -> None:  # pragma: no cover - IO heavy
        super().log_request(code, size)


def serve(app: Flask, host: str, port: int) -> None:
    """Run *app* on *host*:*port* with structured logging."""

    logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")
    logging.getLogger().setLevel(logging.INFO)
    logging.info("Listening on http://%s:%s", host, port)

    server = make_server(host, port, app, request_handler=LoggingWSGIRequestHandler)

    try:
        server.serve_forever()
    except KeyboardInterrupt:  # pragma: no cover - manual interrupt
        logging.info("Shutting down")
    finally:
        server.server_close()


def create_app(config: RemoClipConfig) -> Flask:
    app = Flask(__name__)
    session_factory = create_session_factory(config.server.db_path)
    app.config["SESSION_FACTORY"] = session_factory

    logger = logging.getLogger(__name__)
    allow_deletions = config.server.allow_deletions

    def _seed_clipboard_value() -> str:
        with session_scope(session_factory) as session:
            event = (
                session.query(ClipboardEvent)
                .filter(ClipboardEvent.action.in_(["copy", "paste"]))
                .order_by(ClipboardEvent.timestamp.desc())
                .first()
            )
            if event is not None:
                return event.content
        return ""

    def _create_clipboard_backend() -> ClipboardBackend:
        initial_value = _seed_clipboard_value()
        if config.server.clipboard_backend == "system":
            if is_system_clipboard_available():
                return SystemClipboardBackend()
            warn_if_unavailable(logger, "system")
        return PrivateClipboardBackend(_value=initial_value)

    clipboard_backend = _create_clipboard_backend()
    app.config["CLIPBOARD_BACKEND"] = clipboard_backend

    def _format_timestamp(value: datetime) -> str:
        if value.tzinfo is None:
            value = value.replace(tzinfo=timezone.utc)
        else:
            value = value.astimezone(timezone.utc)
        return value.isoformat().replace("+00:00", "Z")

    def _verify_token() -> Any | None:
        if not config.security_token:
            return None
        header = request.headers.get(SECURITY_TOKEN_HEADER)
        if header != config.security_token:
            return jsonify({"error": "invalid token"}), 401
        return None

    @app.before_request
    def _enforce_token() -> Any | None:
        return _verify_token()

    def _log_event(hostname: str, action: str, content: str) -> None:
        with session_scope(session_factory) as session:
            session.add(
                ClipboardEvent(
                    hostname=hostname,
                    action=action,
                    content=content,
                )
            )

    def _parse_optional_positive_int(value: Any, field: str) -> int | None:
        if value is None:
            return None
        try:
            number = int(value)
        except (TypeError, ValueError) as exc:
            raise ValueError(f"{field} must be an integer") from exc
        if number <= 0:
            raise ValueError(f"{field} must be positive")
        return number

    def _parse_required_positive_int(value: Any, field: str) -> int:
        number = _parse_optional_positive_int(value, field)
        if number is None:
            raise ValueError(f"{field} must be provided")
        return number

    def _validate_payload(data: dict[str, Any], expect_content: bool) -> dict[str, Any]:
        if not data or "hostname" not in data:
            raise ValueError("JSON payload must include 'hostname'")
        if expect_content and "content" not in data:
            raise ValueError("JSON payload must include 'content'")
        return data

    @app.post("/copy")
    def copy_content():
        try:
            data = request.get_json(force=True, silent=False)
            payload = _validate_payload(data, expect_content=True)
            content = str(payload["content"])
            clipboard_backend.copy(content)
            _log_event(str(payload["hostname"]), "copy", content)
            return jsonify({"status": "ok"})
        except Exception as exc:  # pragma: no cover - defensive
            logging.exception("Failed to handle /copy request")
            return jsonify({"error": str(exc)}), 400

    @app.get("/paste")
    def paste_content():
        try:
            data = request.get_json(silent=True) or {}
            payload = _validate_payload(data, expect_content=False)
            event_id = _parse_optional_positive_int(data.get("id"), "id")

            if event_id is not None:
                with session_scope(session_factory) as session:
                    event = (
                        session.query(ClipboardEvent)
                        .filter(
                            ClipboardEvent.id == event_id,
                            ClipboardEvent.action != "history",
                        )
                        .one_or_none()
                    )
                    if event is None:
                        return jsonify({"error": "history entry not found"}), 404
                    content = event.content
            else:
                content = clipboard_backend.paste()
            _log_event(str(payload["hostname"]), "paste", content)
            return jsonify({"content": content})
        except Exception as exc:  # pragma: no cover - defensive
            logging.exception("Failed to handle /paste request")
            return jsonify({"error": str(exc)}), 400

    @app.get("/history")
    def history():
        try:
            data = request.get_json(silent=True) or {}
            payload = _validate_payload(data, expect_content=False)
            limit = _parse_optional_positive_int(data.get("limit"), "limit")
            event_id = _parse_optional_positive_int(data.get("id"), "id")

            with session_scope(session_factory) as session:
                if event_id is not None:
                    event = session.get(ClipboardEvent, event_id)
                    if event is None:
                        return jsonify({"error": "history entry not found"}), 404
                    events = [
                        {
                            "id": event.id,
                            "timestamp": _format_timestamp(event.timestamp),
                            "hostname": event.hostname,
                            "action": event.action,
                            "content": event.content,
                        }
                    ]
                else:
                    query = (
                        session.query(ClipboardEvent)
                        .filter(ClipboardEvent.action != "history")
                        .order_by(ClipboardEvent.timestamp.desc())
                    )
                    if limit is not None:
                        query = query.limit(limit)
                    events = [
                        {
                            "id": item.id,
                            "timestamp": _format_timestamp(item.timestamp),
                            "hostname": item.hostname,
                            "action": item.action,
                            "content": item.content,
                        }
                        for item in query.all()
                    ]
            log_payload: dict[str, Any] = {
                "event_ids": [item["id"] for item in events],
            }
            if limit is not None:
                log_payload["limit"] = limit
            if event_id is not None:
                log_payload["id"] = event_id
            _log_event(
                str(payload["hostname"]),
                "history",
                json.dumps(log_payload),
            )
            return jsonify({"history": events})
        except Exception as exc:  # pragma: no cover - defensive
            logging.exception("Failed to handle /history request")
            return jsonify({"error": str(exc)}), 400

    @app.delete("/history")
    def delete_history():
        try:
            data = request.get_json(force=True, silent=False)
            payload = _validate_payload(data, expect_content=False)
            event_id = _parse_required_positive_int(payload.get("id"), "id")

            if not allow_deletions:
                return jsonify({"error": "history deletions are disabled"}), 403

            with session_scope(session_factory) as session:
                event = session.get(ClipboardEvent, event_id)
                if event is None or event.action == "history":
                    return jsonify({"error": "history entry not found"}), 404
                session.delete(event)
            return jsonify({"status": "deleted"})
        except ValueError as exc:
            return jsonify({"error": str(exc)}), 400
        except Exception as exc:  # pragma: no cover - defensive
            logging.exception("Failed to handle /history delete request")
            return jsonify({"error": str(exc)}), 400

    return app


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Run remoclip HTTP server.",
        epilog=(
            "Configure the clipboard backend in the YAML config file via the "
            "'clipboard_backend' option (system|private)."
        ),
    )
    parser.add_argument(
        "--config",
        default=str(DEFAULT_CONFIG_PATH),
        help="Path to configuration file (default: ~/.remoclip.yaml)",
    )
    args = parser.parse_args()

    config = load_config(args.config)
    app = create_app(config)

    serve(app, config.server.host, config.server.port)


if __name__ == "__main__":  # pragma: no cover
    main()
